from dataclasses import dataclass
from typing import List, Optional, Union, Tuple, Callable
import sqlparse
from sqlparse.sql import IdentifierList, Token, Identifier, Where, Comparison
from functools import partial


@dataclass
class FloatRangeCondition:
    min: Optional[float]
    max: Optional[float]


@dataclass
class WhereClause:
    column_name: str
    condition: Union[FloatRangeCondition]


@dataclass
class Query:
    table_name: str
    columns: Union[List[str], str]
    where_clauses: List[WhereClause]


# def parse_sql_to_query(raw: str) -> Query:
#     parsed = sqlparse.parse(raw)[0]
#     important_tokens = [token for token in parsed if token.is_whitespace is False]
#     from_position = -1
#     for i, token in enumerate(important_tokens):
#         if token.normalized == "FROM":
#             from_position = i
#     table_name: str = important_tokens[from_position + 1].value
#     table_name = table_name.strip("`")
#     where_position = -1
#     for i, token in enumerate(important_tokens):
#         if token.normalized == "WHERE":
#             where_position = i
#
#     return Query(
#         table_name=table_name,
#         columns=[""],
#         where_clauses=[]
#     )


def query(
    columns: Union[List[str], str],
    table_name: str,
    where_clauses: List[WhereClause],
) -> Query:
    return Query(table_name=table_name, columns=columns, where_clauses=where_clauses)


parser = Callable[[partial, List[Token]], Tuple[partial, List[Token]]]


def parse_select(parse: partial, tokens: List[Token]) -> Tuple[partial, List[Token]]:
    if tokens[0].normalized == "SELECT":
        return parse, tokens[1:]
    else:
        raise Exception("SQL query must start with a select")


def parse_columns(parse: partial, tokens: List[Token]) -> Tuple[partial, List[Token]]:
    token = tokens[0]
    if token.normalized == "*":
        return partial(parse, "*"), tokens[1:]
    elif type(token) == Identifier:
        identifier_list = [token.normalized]
        return partial(parse, identifier_list), tokens[1:]
    elif type(token) == IdentifierList:
        identifier_list = [
            tok.normalized for tok in token.tokens if type(tok) == Identifier
        ]
        return partial(parse, identifier_list), tokens[1:]
    else:
        raise Exception("Must specify columns")


def parse_from(parse: partial, tokens: List[Token]) -> Tuple[partial, List[Token]]:
    if tokens[0].normalized == "FROM":
        return parse, tokens[1:]
    else:
        raise Exception("Columns must be followed by a FROM clause")


def parse_table(parse: partial, tokens: List[Token]) -> Tuple[partial, List[Token]]:
    if type(tokens[0]) == Identifier:
        parse = partial(parse, tokens[0].normalized)
        return parse, tokens[1:]
    else:
        raise Exception("From statement must be followed by a table name")


def parse_where(parse: partial, tokens: List[Token]) -> Tuple[partial, List[Token]]:
    next = tokens[0]
    if type(next) == Where:
        where_tokens = next.tokens
        comparisons = [tok for tok in where_tokens if type(tok) == Comparison]
        conditions = []
        for comparison in comparisons:
            column = comparison.left.normalized
            num = float(comparison.right.normalized)
            comparator = comparison.tokens[2].normalized
            if comparator == "<":
                conditions.append(WhereClause(column, FloatRangeCondition(None, num)))
            elif comparator == ">":
                conditions.append(WhereClause(column, FloatRangeCondition(num, None)))
        return partial(parse, conditions), tokens[1:]
    else:
        raise Exception(
            "Where select statement must either terminate or be followed by a where clause"
        )


def parse_sql_to_query(raw: str) -> Query:
    parsed = sqlparse.parse(raw)[0]
    tokens = [token for token in parsed if token.is_whitespace is False]
    parse = partial(query)
    parse, tokens = parse_select(parse, tokens)
    parse, tokens = parse_columns(parse, tokens)
    parse, tokens = parse_from(parse, tokens)
    parse, tokens = parse_table(parse, tokens)
    if len(tokens) == 0:
        parse = partial(parse, [])
        return parse()
    parse, tokens = parse_where(parse, tokens)
    return parse()


@dataclass
class ParseResult:
    success: bool
    query_inputs: List[Token]
    query_result: Optional[Query]


if __name__ == "__main__":
    print(
        parse_sql_to_query(
            "select age, height from synth_user where age > 5 and height < 2"
        )
    )



from dataclasses import dataclass
from typing import List, TypeVar, Callable, Union, Dict

import pandas as pd

from syntheticdb.query_parser import Query, FloatRangeCondition, parse_sql_to_query

T = TypeVar("T")


def sample_until(callable: Callable[[], T], condition: Callable[[T], bool]) -> T:
    while True:
        out_val = callable()
        if condition(out_val):
            return out_val


@dataclass(frozen=True)
class FloatColumn:
    sample: Callable[[], float]
    cdf: Callable[[float], float]

    def where(self, condition: FloatRangeCondition):
        if condition.min and condition.max:

            def new_cdf(point):
                cd_at_min = self.cdf(condition.min)
                cd_at_max = self.cdf(condition.max)
                if point < condition.min:
                    return 0
                if condition.min < point < condition.max:
                    return self.cdf(point) - cd_at_min
                if point > condition.max:
                    return cd_at_max

            return FloatColumn(
                sample=lambda: sample_until(
                    self.sample, lambda x: condition.max > x > condition.min
                ),
                cdf=new_cdf,
            )
        if condition.min:

            def new_cdf(point):
                cd_at_min = self.cdf(condition.min)
                if point < condition.min:
                    return 0
                if condition.min < point:
                    return self.cdf(point) - cd_at_min

            return FloatColumn(
                sample=lambda: sample_until(self.sample, lambda x: x > condition.min),
                cdf=new_cdf,
            )
        if condition.max:

            def new_cdf(point):
                cd_at_max = self.cdf(condition.max)
                if point < condition.max:
                    return self.cdf(point)
                if condition.max < point:
                    return cd_at_max

            return FloatColumn(
                sample=lambda: sample_until(self.sample, lambda x: x < condition.max),
                cdf=new_cdf,
            )
        raise Exception(
            "Invalid condition, must have at least one range value populated"
        )

    def prob(self):
        return self.cdf(1e10)


@dataclass(frozen=True)
class Table:
    columns: Dict[str, Union[FloatColumn]]
    row_count: int

    def get_row_num(self) -> int:
        probability = 1
        for column in self.columns.values():
            probability = probability * column.prob()
        return int(self.row_count * probability)


@dataclass
class DataBase:
    tables: Dict[str, Table]

    def select(self, query_text: str):
        query = parse_sql_to_query(query_text)
        result = self.select_from_query(query)
        return result

    def select_from_query(self, query: Query) -> Dict[str, List[float]]:
        table_name = query.table_name
        where_clauses = query.where_clauses
        table = self.tables.get(table_name.strip("`"))
        view_table = Table(table.columns, table.row_count)
        for clause in where_clauses:
            column = view_table.columns.get(clause.column_name, None)
            if column is None:
                raise Exception(
                    f"column with name: {clause.column_name}, does not exist"
                )
            if type(clause.condition) not in [FloatRangeCondition]:
                raise Exception(f"Condition type not supported")
            if type(clause.condition) == FloatRangeCondition:
                view_table.columns[clause.column_name] = column.where(clause.condition)
        new_row_count = view_table.get_row_num()
        columns_to_return = {}
        if query.columns == "*":
            for name, col in view_table.columns.items():
                columns_to_return[name] = [col.sample() for _ in range(new_row_count)]
        else:
            for name, col in view_table.columns.items():
                if name in query.columns:
                    columns_to_return[name] = [
                        col.sample() for _ in range(new_row_count)
                    ]
        return pd.DataFrame.from_dict(columns_to_return)
